# Generated by Django 5.2.1 on 2025-06-10 20:21

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("axes", "0009_add_session_hash"),
    ]

    operations = [
        migrations.CreateModel(
            name="AccessAttemptExpiration",
            fields=[
                (
                    "access_attempt",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="expiration",
                        serialize=False,
                        to="axes.accessattempt",
                        verbose_name="Access Attempt",
                    ),
                ),
                (
                    "expires_at",
                    models.DateTimeField(
                        help_text="The time when access attempt expires and is no longer valid.",
                        verbose_name="Expires At",
                    ),
                ),
            ],
            options={
                "verbose_name": "access attempt expiration",
                "verbose_name_plural": "access attempt expirations",
            },
        ),
    ]
